import typer

from ..cli.utils import parse_dict_or_list_param
from ..mqtt import MQTTClient

app = typer.Typer(help="MQTT management commands")


@app.command()
def start_broker(
    host: str = "localhost",
    port: int = 1883,
    username: str = None,
    password: str = None,
    log_level: str = "INFO",
):
    """Start an MQTT broker."""
    MQTTClient.start_broker(host, port, username, password, log_level)



# If you wanted to use another OpenTelemetry destination such as the open-source Jaeger,
# setup the container locally and use the following code

from opentelemetry import trace
from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.sdk.trace import Resource, TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

# Add more open telemetry exporters here


def init_tracer(
    name: str,
    host: str = "localhost",
    port: int = 6831,
):
    jaeger_exporter = JaegerExporter(
        agent_host_name=host,  # Replace with your Jaeger agent host
        agent_port=port,  # Replace with your Jaeger agent port
    )

    span_processor = SimpleSpanProcessor(jaeger_exporter)
    provider = TracerProvider(
        active_span_processor=span_processor,
        resource=Resource.create({"service.name": f"flowerpower.{name}"}),
    )
    trace.set_tracer_provider(provider)

    return trace



import importlib.util

import typer
from loguru import logger

# Import your existing pipeline functions
# from ..pipeline import (
#     add as add_pipeline_,
#     add_job as add_pipeline_job_,
#     all_pipelines as all_pipelines_,
#     delete as delete_pipeline_,
#     get_summary as get_pipeline_summary_,
#     new as new_pipeline_,
#     run as run_pipeline_,
#     run_job as run_pipeline_job_,
#     schedule as schedule_pipeline_,
#     save_dag as save_pipeline_dag_,
#     show_dag as show_pipeline_dag_,
#     show_summary as show_pipeline_summary_,
#     # start_mqtt_listener as start_mqtt_listener_,
# )
from ..pipeline import Pipeline, PipelineManager
from .utils import parse_dict_or_list_param, parse_param_dict

# Optional imports
if importlib.util.find_spec("apscheduler"):
    from ..scheduler import get_schedule_manager
    from ..scheduler import start_worker as start_worker_
else:
    get_schedule_manager = None
    start_worker_ = None


app = typer.Typer(help="Pipeline management commands")


@app.command()
def run(
    name: str,
    executor: str = "local",
    base_dir: str | None = None,
    inputs: str | None = None,
    final_vars: str | None = None,
    config: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    reload: bool = False,
    storage_options: str | None = None,
):
    """
    Run the specified pipeline.

    Args:
        name: Name of the pipeline to run
        executor: Executor to use
        base_dir: Base directory for the pipeline
        inputs: Input parameters as JSON, dict string, or key=value pairs
        final_vars: Final variables as JSON or list
        config: Config for the hamilton pipeline executor
        with_tracker: Enable tracking with hamilton ui
        with_opentelemetry: Enable OpenTelemetry tracing
        reload: Reload pipeline before running
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    # JSON inputs
    pipeline run my_pipeline --inputs '{"key": "value"}'

    # Dict string inputs
    pipeline run my_pipeline --inputs "{'key': 'value'}"

    # Key-value pair inputs
    pipeline run my_pipeline --inputs 'key1=value1,key2=value2'

    # List final vars
    pipeline run my_pipeline --final-vars '["var1", "var2"]'

    # Storage options
    pipeline run my_pipeline --storage-options 'endpoint=http://localhost,use_ssl=true'
    """
    parsed_inputs = parse_dict_or_list_param(inputs, "dict")
    parsed_config = parse_dict_or_list_param(config, "dict")
    parsed_final_vars = parse_dict_or_list_param(final_vars, "list")
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        pipeline.run(
            executor=executor,
            inputs=parsed_inputs,
            final_vars=parsed_final_vars,
            config=parsed_config,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
        )


@app.command()
def run_job(
    name: str,
    executor: str = "local",
    base_dir: str | None = None,
    inputs: str | None = None,
    final_vars: str | None = None,
    config: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    reload: bool = False,
    storage_options: str | None = None,
):
    """
    Run the specified pipeline job.

    Args:
        name: Name of the pipeline job to run
        executor: Executor to use
        base_dir: Base directory for the pipeline
        inputs: Input parameters as JSON, dict string, or key=value pairs
        final_vars: Final variables as JSON or list
        config: Config for the hamilton pipeline executor
        with_tracker: Enable tracking with hamilton ui
        with_opentelemetry: Enable OpenTelemetry tracing
        reload: Reload pipeline before running
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    # JSON inputs
    pipeline run-job 123 --inputs '{"key": "value"}'

    # Dict string inputs
    pipeline run-job 123 --inputs "{'key': 'value'}"

    # Key-value pair inputs
    pipeline run-job 123 --inputs 'key1=value1,key2=value2'

    # List final vars
    pipeline run-job 123 --final-vars '["var1", "var2"]'

    # Storage options
    pipeline run-job 123 --storage-options 'endpoint=http://localhost,use_ssl=true'
    """
    parsed_inputs = parse_dict_or_list_param(inputs, "dict")
    parsed_config = parse_dict_or_list_param(config, "dict")
    parsed_final_vars = parse_dict_or_list_param(final_vars, "list")
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        pipeline.run_job(
            executor=executor,
            inputs=parsed_inputs,
            final_vars=parsed_final_vars,
            config=parsed_config,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
        )


@app.command()
def add_job(
    name: str,
    executor: str = "local",
    base_dir: str | None = None,
    inputs: str | None = None,
    final_vars: str | None = None,
    config: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    reload: bool = False,
    storage_options: str | None = None,
):
    """
    Add a job to the specified pipeline.

    Args:
        name: Name of the pipeline to add as job
        executor: Executor to use
        base_dir: Base directory for the pipeline
        inputs: Input parameters as JSON, dict string, or key=value pairs
        final_vars: Final variables as JSON or list
        config: Config for the hamilton pipeline executor
        with_tracker: Enable tracking with hamilton ui
        with_opentelemetry: Enable OpenTelemetry tracing
        reload: Reload pipeline before running
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    # JSON inputs
    pipeline add-job my_pipeline --inputs '{"key": "value"}'

    # Dict string inputs
    pipeline add-job my_pipeline --inputs "{'key': 'value'}"

    # Key-value pair inputs
    pipeline add-job my_pipeline --inputs 'key1=value1,key2=value2'

    # List final vars
    pipeline add-job my_pipeline --final-vars '["var1", "var2"]'

    # Storage options
    pipeline add-job my_pipeline --storage-options 'endpoint=http://localhost,use_ssl=true'
    """
    parsed_inputs = parse_dict_or_list_param(inputs, "dict")
    parsed_config = parse_dict_or_list_param(config, "dict")
    parsed_final_vars = parse_dict_or_list_param(final_vars, "list")
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    storage_options = (parsed_storage_options or {},)

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        pipeline.add_job(
            executor=executor,
            inputs=parsed_inputs,
            final_vars=parsed_final_vars,
            config=parsed_config,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            reload=reload,
        )


@app.command()
def schedule(
    name: str,
    executor: str = "local",
    base_dir: str | None = None,
    trigger_type: str = "cron",
    inputs: str | None = None,
    final_vars: str | None = None,
    config: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    paused: bool = False,
    coalesce: str = "latest",
    misfire_grace_time: float | None = None,
    max_jitter: float | None = None,
    max_running_jobs: int | None = None,
    conflict_policy: str = "do_nothing",
    crontab: str | None = None,
    cron_params: str | None = None,
    interval_params: str | None = None,
    calendarinterval_params: str | None = None,
    date_params: str | None = None,
    storage_options: str | None = None,
    overwrite: bool = False,
):
    """
    Schedule a pipeline with various configuration options.

    Args:
        name: Name of the pipeline to schedule
        executor: Executor to use
        base_dir: Base directory for the pipeline
        trigger_type: Type of schedule
        inputs: Input parameters as JSON, dict string, or key=value pairs
        final_vars: Final variables as JSON or list
        config: Config for the hamilton pipeline executor
        with_tracker: Enable tracking with hamilton ui
        with_opentelemetry: Enable OpenTelemetry tracing
        paused: Start the job in paused state
        coalesce: Coalesce policy
        misfire_grace_time: Misfire grace time
        max_jitter: Maximum jitter
        max_running_jobs: Maximum running jobs
        conflict_policy: Conflict policy
        crontab: Crontab expression
        cron_params: Cron parameters as JSON or key=value pairs
        interval_params: Interval parameters as JSON or key=value pairs
        calendarinterval_params: Calendar interval parameters as JSON or key=value pairs
        date_params: Date parameters as JSON or key=value pairs
        storage_options: Storage options as JSON, dict string, or key=value pairs
        overwrite: Overwrite existing schedule

    Examples:
    # JSON inputs
    pipeline schedule my_pipeline --inputs '{"key": "value"}'

    # Dict string inputs
    pipeline schedule my_pipeline --inputs "{'key': 'value'}"

    # Key-value pair inputs
    pipeline schedule my_pipeline --inputs 'key1=value1,key2=value2'

    # List final vars
    pipeline schedule my_pipeline --final-vars '["var1", "var2"]'

    # Storage options
    pipeline schedule my_pipeline --storage-options 'endpoint=http://localhost,use_ssl=true'

    # Cron schedule
    pipeline schedule my_pipeline --trigger-type cron --crontab '0 0 * * *'

    # Interval schedule
    pipeline schedule my_pipeline --trigger-type interval --interval_params minutes=1

    # Calendar interval schedule
    pipeline schedule my_pipeline --trigger-type calendarinterval --calendarinterval_params month=5

    # Date schedule
    pipeline schedule my_pipeline --trigger-type date --date_params run_date='2021-01-01 12:00:01'

    """
    if get_schedule_manager is None:
        raise ValueError("APScheduler not installed. Please install it first.")

    # Parse inputs
    parsed_inputs = parse_dict_or_list_param(inputs, "dict")
    parsed_config = parse_dict_or_list_param(config, "dict")
    parsed_final_vars = parse_dict_or_list_param(final_vars, "list")
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    # Parse various parameter dictionaries
    cron_params_dict = parse_param_dict(cron_params)
    interval_params_dict = parse_param_dict(interval_params)
    calendarinterval_params_dict = parse_param_dict(calendarinterval_params)
    date_params_dict = parse_param_dict(date_params)

    # Combine all parameter dictionaries
    kwargs = {
        **cron_params_dict,
        **interval_params_dict,
        **calendarinterval_params_dict,
        **date_params_dict,
    }

    # Add crontab if provided
    if crontab is not None:
        kwargs["crontab"] = crontab

    # Convert numeric parameters
    for key in ["weeks", "days", "hours", "minutes", "seconds"]:
        if key in kwargs:
            try:
                kwargs[key] = float(kwargs[key])
            except ValueError:
                logger.warning(f"Could not convert {key} to float: {kwargs[key]}")

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        id_ = pipeline.schedule(
            executor=executor,
            trigger_type=trigger_type,
            inputs=parsed_inputs,
            final_vars=parsed_final_vars,
            config=parsed_config,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            paused=paused,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            max_jitter=max_jitter,
            max_running_jobs=max_running_jobs,
            conflict_policy=conflict_policy,
            overwrite=overwrite,
            **kwargs,
        )

    logger.info(f"Job {id_} scheduled.")


@app.command()
def schedule_all(
    executor: str = "local",
    base_dir: str | None = None,
    inputs: str | None = None,
    final_vars: str | None = None,
    config: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    paused: bool = False,
    coalesce: str = "latest",
    misfire_grace_time: float | None = None,
    max_jitter: float | None = None,
    max_running_jobs: int | None = None,
    conflict_policy: str = "do_nothing",
    storage_options: str | None = None,
    overwrite: bool = False,
):
    """
    Schedule all pipelines using the pipeline specific configurations (`conf/pipelines/<name>.yml`).

    Args:
        executor: Executor to use
        base_dir: Base directory for the pipeline
        inputs: Input parameters as JSON, dict string, or key=value pairs
        final_vars: Final variables as JSON or list
        config: Config for the hamilton pipeline executor
        with_tracker: Enable tracking with hamilton ui
        with_opentelemetry: Enable OpenTelemetry tracing
        paused: Start the job in paused state
        coalesce: Coalesce policy
        misfire_grace_time: Misfire grace time
        max_jitter: Maximum jitter
        max_running_jobs: Maximum running jobs
        conflict_policy: Conflict policy
        storage_options: Storage options as JSON, dict string, or key=value pairs
        overwrite: Overwrite existing schedule

    Examples:
    pipeline schedule-all
    """
    if get_schedule_manager is None:
        raise ValueError("APScheduler not installed. Please install it first.")

    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with PipelineManager(
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as manager:
        manager.schedule_all(
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            config=config,
            with_tracker=with_tracker,
            with_opentelemetry=with_opentelemetry,
            paused=paused,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            max_jitter=max_jitter,
            max_running_jobs=max_running_jobs,
            conflict_policy=conflict_policy,
            overwrite=overwrite,
        )


@app.command()
def new(
    name: str,
    base_dir: str | None = None,
    storage_options: str | None = None,
    overwrite: bool = False,
):
    """
    Create a new pipeline.

    Args:
        name: Name of the new pipeline
        base_dir: Base directory for the new pipeline
        overwrite: Overwrite existing pipeline
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    pipeline new my_pipeline
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with PipelineManager(
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as manager:
        manager.new(name=name, overwrite=overwrite)


@app.command()
def delete(
    name: str,
    base_dir: str | None = None,
    cfg: bool = False,
    module: bool = False,
    storage_options: str | None = None,
):
    """
    Delete the specified pipeline.

    Args:
        name: Name of the pipeline to delete
        base_dir: Base directory for the pipeline
        cfg: Remove associated configuration
        module: Remove associated module
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    pipeline delete my_pipeline
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        pipeline.delete(cfg=cfg, module=module)


@app.command()
def show_dag(
    name: str, base_dir: str | None = None, storage_options: str | None = None
):
    """
    Show the DAG of the specified pipeline.

    Args:
        name: Name of the pipeline to show
        base_dir: Base directory for the pipeline
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    pipeline show-dag my_pipeline
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        pipeline.show_dag()


@app.command()
def save_dag(
    name: str, base_dir: str | None = None, storage_options: str | None = None
):
    """
    Save the DAG of the specified pipeline.

    Args:
        name: Name of the pipeline to save
        base_dir: Base directory for the pipeline
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    pipeline save-dag my_pipeline
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with Pipeline(
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as pipeline:
        pipeline.save_dag()


@app.command()
def show_pipelines(base_dir: str | None = None, storage_options: str | None = None):
    """
    List all available pipelines.

    Args:
        base_dir: Base directory for the pipelines
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    pipeline list-pipelines
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")
    with PipelineManager(
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as manager:
        manager.show_pipelines()


@app.command()
def show_summary(
    name: str | None = None,
    cfg: bool = True,
    module: bool = True,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Show the summary of the specified pipeline.

    Args:
        name: Name of the pipeline to show
        base_dir: Base directory for the pipeline
        cfg: Show configuration
        module: Show module information
        storage_options: Storage options as JSON, dict string, or key=value pairs

    Examples:
    pipeline show-summary my_pipeline
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    with PipelineManager(
        base_dir=base_dir,
        storage_options=parsed_storage_options or {},
    ) as manager:
        manager.show_summary(name=name, cfg=cfg, module=module)



from munch import Munch, munchify
from pydantic import Field

from ..base import BaseConfig


class PipelineRunConfig(BaseConfig):
    final_vars: list[str] = Field(default_factory=list)
    inputs: dict | Munch = Field(default_factory=dict)
    executor: str | None = None
    config: dict | Munch = Field(default_factory=dict)
    with_tracker: bool = False
    with_opentelemetry: bool = False

    def model_post_init(self, __context):
        if isinstance(self.inputs, dict):
            self.inputs = munchify(self.inputs)



import posixpath

from sanic import Blueprint
from sanic.exceptions import SanicException
from sanic.response import json
from sanic_ext import openapi, validate

from ...cfg import PipelineConfig, ProjectConfig

bp = Blueprint("api_flowerpower_cfg", url_prefix="api/cfg")


@bp.get("/project")
async def get_project(request) -> json:
    cfg = request.app.ctx.pipeline_manager.cfg.project.to_dict()
    # cfg.pop("fs")
    return json({"cfg": cfg})


@bp.post("/project")
@validate(json=ProjectConfig)
async def update_project(request, body: ProjectConfig) -> json:
    cfg = request.app.ctx.pipeline_manager.cfg.copy()
    cfg.project.update(body.model_dump())
    try:
        cfg.save()
    except NotImplementedError as e:
        raise SanicException(f"Update failed. {e}", status_code=404)
    return json({"cfg": cfg})


@bp.get("/pipeline/<pipeline_name>")
async def get_pipeline(request, pipeline_name) -> json:
    if pipeline_name != request.app.ctx.pipeline_manager.cfg.pipeline.name:
        request.app.ctx.pipeline_manager.load_config(pipeline_name)
    cfg = request.app.ctx.pipeline_manager.cfg.pipeline.to_dict()
    return json({"cfg": cfg})


@bp.post("/pipeline/<pipeline_name>")
@validate(json=PipelineConfig)
async def update_pipeline(request, pipeline_name, body: PipelineConfig) -> json:
    if pipeline_name != request.app.ctx.pipeline_manager.cfg.pipeline.name:
        request.app.ctx.pipeline_manager.load_config(pipeline_name)
    cfg = request.app.ctx.pipeline_manager.cfg.copy()
    cfg.pipeline.update(body.model_dump())
    try:
        cfg.save()
    except NotImplementedError as e:
        raise SanicException(f"Update failed. {e}", status_code=404)

    return json({"cfg": cfg})



import os

from dotenv import load_dotenv
from orjson import dumps, loads
from sanic import Sanic

# from sanic.response import json

from .setup import setup
from ..cli.utils import parse_dict_or_list_param

load_dotenv()


def create_app(args):
    # Note: args is the parsed command line arguments when the server is
    # started using sanic build-in server. e.g.
    # sanic src.flowerpower.http.main:create_app -factory --base-dir /path/to/base_dir
    # app = setup(args.base_dir)
    base_dir = (
        args.base_dir
        if hasattr(args, "base_dir") or os.getenv("FLOWERPOWER_BASE_DIR")
        else None
    )

    storage_options = (
        args.storage_options
        if hasattr(args, "storage_options") or os.getenv("FLOWERPOWER_STORAGE_OPTIONS")
        else None
    )
    storage_options = parse_dict_or_list_param(storage_options, param_type="dict") or {}

    # cfg_dir = args.cfg_dir if hasattr(args, "cfg_dir") else "conf"
    # pipelines_dir = (
    #    args.pipelines_dir if hasattr(args, "pipelines_dir") else "pipelines"
    # )

    app = Sanic("flowerpower", dumps=dumps, loads=loads)
    setup(
        app=app,
        base_dir=base_dir,
        storage_options=storage_options,
        # cfg_dir=cfg_dir,
        # pipelines_dir=pipelines_dir,
    )

    # @app.get("/")
    # async def t2(request):
    #     print(request.args)
    #     print(request.json)
    #     return json({"status": "success", "message": "Welcome to FlowerPower"})

    # @app.post("/")
    # async def t1(request):
    #     print(request.args)
    #     print(request.args.get("name"))
    #     print(request.json)
    #     return json({"status": "success", "message": "Welcome to FlowerPower"})

    # # @app.get("/health")
    # # async def health(request):
    # #     return json({"status": "success", "message": "Healthy"})

    # @app.get("/health/<name>")
    # async def health(request, name):
    #     if name is None:
    #         return json({"status": "success", "message": "Healthy"})
    #     return json({"status": "success", "message": f"Healthy {name}"})

    return app



from typing import Any

import datafusion
import duckdb
import orjson
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as pds
from pydantic import BaseModel

from ...utils.sql import sql2polars_filter


class MQTTLoader(BaseModel):
    payload: bytes | dict[str, Any]
    topic: str | None = None
    conn: duckdb.DuckDBPyConnection | None = None
    ctx: datafusion.SessionContext | None = None

    def model_post_init(self, __context):
        if isinstance(self.payload, bytes):
            self.payload = orjson.loads(self.payload)

    def to_pyarrow_table(self) -> pa.Table:
        return pa.Table.from_pydict(self.payload)

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.payload)

    def _to_polars_dataframe(self) -> pl.DataFrame:
        return pl.DataFrame(self.payload)

    def _to_polars_lazyframe(self) -> pl.LazyFrame:
        return pl.LazyFrame(self.payload)

    def to_poars(self, lazy: bool = False) -> pl.DataFrame | pl.LazyFrame:
        if lazy:
            return self._to_polars_lazyframe()
        else:
            return self._to_polars_dataframe()

    def to_duckdb(self, conn: duckdb.DuckDBPyConnection | None = None):
        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn
        return self.conn.from_arrow(self.to_pyarrow_table())

    def to_dataset(self, **kwargs) -> pds.Dataset:
        return pds.dataset(self.to_pyarrow_table(), **kwargs)

    def register_in_duckdb(
        self,
        conn: duckdb.DuckDBPyConnection | None = None,
        name: str | None = None,
    ):
        if name is None:
            name = f"mqtt:{self.topic}"

        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        self.conn.register(name, self.to_pyarrow_table())

    def register_in_datafusion(
        self,
        ctx: datafusion.SessionContext | None = None,
        name: str | None = None,
    ) -> None:
        if name is None:
            name = f"mqtt:{self.topic}"

        if self.ctx is None:
            if ctx is None:
                ctx = datafusion.SessionContext()
            self.ctx = ctx

        ctx.register(name, [self.to_pyarrow_table()])

        return ctx

    def filter(self, filter_expr: str | pl.Expr) -> pl.DataFrame | pl.LazyFrame:
        self._data = self.to_poars()

        pl_schema = (
            self._data.schema
            if isinstance(self._data, pl.DataFrame)
            else self._data.collect_schema()
        )
        filter_expr = (
            sql2polars_filter(filter_expr, pl_schema)
            if isinstance(filter_expr, str)
            else filter_expr
        )
        return self._data.filter(filter_expr)



from ..base import BaseDatasetWriter


class PydalaDatasetWriter(BaseDatasetWriter):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "parquet"
        self.is_pydala_dataset = True



import posixpath
from typing import Any, Generator

import datafusion
import duckdb
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
from fsspec import AbstractFileSystem
from fsspec.utils import get_protocol
from pydantic import BaseModel, ConfigDict

from ..fs import get_filesystem
from ..fs.ext import path_to_glob
from ..utils.polars import pl
from ..utils.sql import sql2polars_filter, sql2pyarrow_filter
from ..fs.storage_options import (
    AwsStorageOptions,
    AzureStorageOptions,
    GcsStorageOptions,
    GitHubStorageOptions,
    GitLabStorageOptions,
    StorageOptions,
)
import importlib

if importlib.util.find_spec("pydala"):
    from pydala.dataset import ParquetDataset
else:
    ParquetDataset = None


class BaseFileIO(BaseModel):
    """
    Base class for file I/O operations supporting various storage backends.
    This class provides a foundation for file operations across different storage systems
    including AWS S3, Google Cloud Storage, Azure Blob Storage, GitHub, and GitLab.

    Args:
        path (str | list[str]): Path or list of paths to file(s).
        storage_options (AwsStorageOptions | GcsStorageOptions | AzureStorageOptions |
                             GitHubStorageOptions | GitLabStorageOptions | dict[str, Any] |  None, optional):
            Storage-specific options for accessing remote filesystems.
        fs (AbstractFileSystem, optional): Filesystem instance for handling file operations.
        format (str, optional): File format extension (without dot).

    Notes:
        ```python
        file_io = BaseFileIO(
            path="s3://bucket/path/to/files",
            storage_options=AwsStorageOptions(
                key="access_key",
                secret="secret_key"
        files = file_io.list_files()
        ```
    Notes:
        - Supports multiple cloud storage backends through different storage options
        - Automatically handles filesystem initialization based on path protocol
        - Supports both single path and multiple path inputs
        - Can read credentials from environment variables when using from_env() methods

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    path: str | list[str]
    storage_options: (
        StorageOptions
        | AwsStorageOptions
        | AzureStorageOptions
        | GcsStorageOptions
        | GitLabStorageOptions
        | GitHubStorageOptions
        | dict[str, Any]
        | None
    ) = None
    fs: AbstractFileSystem | None = None
    format: str | None = None

    def model_post_init(self, __context):
        self._raw_path = self.path
        if isinstance(self.storage_options, dict):
            if "protocol" not in self.storage_options:
                self.storage_options["protocol"] = get_protocol(self.path)
            self.storage_options = StorageOptions(
                **self.storage_options
            ).storage_options
        if isinstance(self.storage_options, StorageOptions):
            self.storage_options = self.storage_options.storage_options

        if self.fs is None:
            self.fs = get_filesystem(
                path=self.path if isinstance(self.path, str) else self.path[0],
                storage_options=self.storage_options,
                fs=self.fs,
                dirfs=True,
            )

        if hasattr(self.storage_options, "protocol"):
            protocol = self.storage_options.protocol
        else:
            protocol = self.fs.protocol
            if protocol == "dir":
                protocol = (
                    self.fs.fs.protocol
                    if isinstance(self.fs.fs.protocol, str)
                    else self.fs.fs.protocol[0]
                )
            if isinstance(protocol, list | tuple):
                protocol = protocol[0]

        if isinstance(self.path, str):
            self.path = (
                self.path.replace(protocol + "://", "")
                .replace(f"**/*.{self.format}", "")
                .replace("**", "")
                .replace("*", "")
                .rstrip("/")
            )

    @property
    def _path(self):
        if self.fs.protocol == "dir":
            if isinstance(self.path, list):
                return [
                    p.replace(self.fs.path.lstrip("/"), "").lstrip("/")
                    for p in self.path
                ]
            else:
                return self.path.replace(self.fs.path.lstrip("/"), "").lstrip("/")
        return self.path

    @property
    def _glob_path(self):
        return path_to_glob(self._path, self.format)

    def list_files(self):
        if isinstance(self._path, list):
            return self._path

        return self.fs.glob(self._glob_path)


class BaseFileLoader(BaseFileIO):
    """
    Base class for file loading operations supporting various file formats.
    This class provides a foundation for file loading operations across different file formats
    including CSV, Parquet, JSON, Arrow, and IPC.

    Args:
        path (str | list[str]): Path or list of paths to file(s).
        format (str, optional): File format extension (without dot).
        fs (AbstractFileSystem, optional): Filesystem instance for handling file operations.
        include_file_path (bool, optional): Include file path in the output DataFrame.
        concat (bool, optional): Concatenate multiple files into a single DataFrame.
        conn (duckdb.DuckDBPyConnection, optional): DuckDB connection instance.
        ctx (datafusion.SessionContext, optional): DataFusion session context instance.

    Examples:
        ```python
        file_loader = BaseFileLoader(
            path="s3://bucket/path/to/files",
            format="csv",
            include_file_path=True,
            concat=True,
            conn=duckdb.connect(),
            ctx=datafusion.SessionContext()
        data = file_loader.to_polars()
        ```
    Notes:
        - Supports multiple file formats including CSV, Parquet, JSON, Arrow, and IPC
        - Automatically handles filesystem initialization based on path protocol
        - Supports both single path and multiple path inputs
        - Supports loading data into DuckDB and DataFusion for SQL operations

    """

    include_file_path: bool = False
    concat: bool = True
    batch_size: int | None = None
    conn: duckdb.DuckDBPyConnection | None = None
    ctx: datafusion.SessionContext | None = None
    jsonlines: bool | None = None
    partitioning: str | list[str] | pds.Partitioning | None = None

    def load(self, **kwargs):
        reload = kwargs.pop("reload", False)
        if "include_file_path" in kwargs:
            if self.include_file_path != kwargs["include_file_path"]:
                reload = True
                self.include_file_path = kwargs.pop("include_file_path")
            else:
                kwargs.pop("include_file_path")

        if "concat" in kwargs:
            if self.concat != kwargs["concat"]:
                reload = True
                self.concat = kwargs.pop("concat")
            else:
                kwargs.pop("concat")

        if "batch_size" in kwargs:
            if self.batch_size != kwargs["batch_size"]:
                reload = True
                self.batch_size = kwargs.pop("batch_size")
            else:
                kwargs.pop("batch_size")

        if "partitioning" in kwargs:
            if self.partitioning != kwargs["partitioning"]:
                reload = True
                self.partitioning = kwargs.pop("partitioning")
            else:
                kwargs.pop("partitioning")

        if not hasattr(self, "_data") or self._data is None or reload:
            self._data = self.fs.read_files(
                path=self._glob_path,
                format=self.format,
                include_file_path=self.include_file_path,
                concat=self.concat,
                jsonlines=self.jsonlines or None,
                batch_size=self.batch_size,
                partitioning=self.partitioning,
                **kwargs,
            )

        # else:
        #     reload = kwargs.pop("reload", False)
        #     if self.include_file_path and not "file_path" not in self._data.columns:
        #         reload = True
        #     if isinstance(self._data, pl.DataFrame | pa.Table) and not self.concat:
        #         reload = True
        #     if isinstance(self._data, Generator):
        #         reload = True

        #     if reload:
        #         self._data = self.fs.read_files(
        #             path=self._glob_path,
        #             format=self.format,
        #             include_file_path=self.include_file_path,
        #             concat=self.concat,
        #             jsonlines=self.jsonlines or None,
        #             batch_size=self.batch_size,
        #             **kwargs,
        #         )

    def to_pandas(self, **kwargs) -> pd.DataFrame | list[pd.DataFrame]:
        """Convert data to Pandas DataFrame(s).

        Returns:
            pd.DataFrame | list[pd.DataFrame]: Pandas DataFrame or list of DataFrames.
        """
        self.load(**kwargs)

    def iter_pandas(
        self, batch_size: int = 1, **kwargs
    ) -> Generator[pd.DataFrame, None, None]:
        """Iterate over Pandas DataFrames.

        Returns:
            Generator[pd.DataFrame, None, None]: Generator of Pandas DataFrames.
        """
        self.load(batch_size=batch_size, **kwargs)
        if isinstance(self._data, list | Generator):
            for df in self._data:
                yield df if isinstance(df, pd.DataFrame) else df.to_pandas()
        else:
            yield (
                self._data
                if isinstance(self._data, pd.DataFrame)
                else self._data.to_pandas()
            )

    def _to_polars_dataframe(self, **kwargs) -> pl.DataFrame | list[pl.DataFrame]:
        """Convert data to Polars DataFrame(s).

        Returns:
            pl.DataFrame | list[pl.DataFrame]: Polars DataFrame or list of DataFrames.
        """
        self.load(**kwargs)
        if isinstance(self._data, list):
            df = [
                df if isinstance(self._data, pl.DataFrame) else pl.from_arrow(df)
                for df in self._data
            ]
            return pl.concat(df) if self.concat else df
        return (
            self._data
            if isinstance(self._data, pl.DataFrame)
            else pl.from_arrow(self._data)
        )

    def _iter_polars_dataframe(
        self, batch_size: int = 1, **kwargs
    ) -> Generator[pl.DataFrame, None, None]:
        """Iterate over Polars DataFrames.

        Returns:
            Generator[pl.DataFrame, None, None]: Generator of Polars DataFrames.
        """
        self.load(batch_size=batch_size, **kwargs)
        if isinstance(self._data, list | Generator):
            for df in self._data:
                yield df if isinstance(df, pl.DataFrame) else pl.from_arrow(df)
        else:
            yield (
                self._data
                if isinstance(self._data, pl.DataFrame)
                else pl.from_arrow(self._data)
            )

    def _to_polars_lazyframe(self, **kwargs) -> pl.LazyFrame | list[pl.LazyFrame]:
        """Convert data to Polars LazyFrame(s).

        Returns:
            pl.LazyFrame | list[pl.LazyFrame]: Polars LazyFrame or list of LazyFrames.
        """
        self.load(**kwargs)
        if not self.concat:
            return [df.lazy() for df in self._to_polars_dataframe()]
        return self._to_polars_dataframe.lazy()

    def _iter_polars_lazyframe(
        self, batch_size: int = 1, **kwargs
    ) -> Generator[pl.LazyFrame, None, None]:
        """Iterate over Polars LazyFrames.

        Returns:
            Generator[pl.LazyFrame, None, None]: Generator of Polars LazyFrames.
        """
        self.load(batch_size=batch_size, **kwargs)
        if isinstance(self._data, list | Generator):
            for df in self._data:
                yield (
                    df.lazy()
                    if isinstance(df, pl.DataFrame)
                    else pl.from_arrow(df).lazy()
                )
        else:
            yield (
                self._data.lazy()
                if isinstance(self._data, pl.DataFrame)
                else pl.from_arrow(self._data).lazy()
            )

    def to_polars(
        self,
        lazy: bool = False,
        **kwargs,
    ) -> pl.DataFrame | pl.LazyFrame | list[pl.DataFrame] | list[pl.LazyFrame]:
        """Convert data to Polars DataFrame or LazyFrame.

        Args:
            lazy (bool, optional): Return a LazyFrame if True, else a DataFrame.

        Returns:
            pl.DataFrame | pl.LazyFrame | list[pl.DataFrame] | list[pl.LazyFrame]: Polars DataFrame or LazyFrame.

        """
        if lazy:
            return self._to_polars_lazyframe(**kwargs)
        return self._to_polars_dataframe(**kwargs)

    def iter_polars(
        self,
        lazy: bool = False,
        batch_size: int = 1,
        **kwargs,
    ) -> Generator[pl.DataFrame | pl.LazyFrame, None, None]:
        if lazy:
            yield from self._iter_polars_lazyframe(batch_size=batch_size, **kwargs)
        yield from self._iter_polars_dataframe(batch_size=batch_size, **kwargs)

    def to_pyarrow_table(self, **kwargs) -> pa.Table | list[pa.Table]:
        """Convert data to PyArrow Table(s).

        Returns:
            pa.Table | list[pa.Table]: PyArrow Table or list of Tables.
        """
        self.load(**kwargs)
        if isinstance(self._data, list):
            df = [
                df.to_arrow(**kwargs) if isinstance(df, pl.DataFrame) else df
                for df in self._data
            ]
            return pa.concat_tables(df) if self.concat else df
        return (
            self._data.to_arrow(**kwargs)
            if isinstance(self._data, pl.DataFrame)
            else self._data
        )

    def iter_pyarrow_table(
        self, batch_size: int = 1, **kwargs
    ) -> Generator[pa.Table, None, None]:
        """Iterate over PyArrow Tables.

        Returns:
            Generator[pa.Table, None, None]: Generator of PyArrow Tables.
        """
        self.load(batch_size=batch_size, **kwargs)
        if isinstance(self._data, list | Generator):
            for df in self._data:
                yield df.to_arrow(**kwargs) if isinstance(df, pl.DataFrame) else df
        else:
            yield (
                self._data.to_arrow(**kwargs)
                if isinstance(self._data, pl.DataFrame)
                else self._data
            )

    def to_duckdb_relation(
        self, conn: duckdb.DuckDBPyConnection | None = None, **kwargs
    ) -> duckdb.DuckDBPyRelation:
        """Convert data to DuckDB relation.

        Args:
            conn (duckdb.DuckDBPyConnection, optional): DuckDB connection instance.

        Returns:

            duckdb.DuckDBPyRelation: DuckDB relation.
        """
        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if hasattr(self, "_data"):
            return self.conn.from_arrow(self.to_pyarrow_table(concat=True))

        else:
            self.load(**kwargs)
            return self.conn.from_arrow(self.to_pyarrow_table(concat=True))

    def register_in_duckdb(
        self, conn: duckdb.DuckDBPyConnection, name: str | None = None, **kwargs
    ) -> None:
        """Register data in DuckDB.

        Args:
            conn (duckdb.DuckDBPyConnection): DuckDB connection instance.

        Returns:
            None
        """
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if hasattr(self, "_data"):
            self.conn.register(name, self.to_pyarrow_table(concat=True))
        else:
            self.conn.register(name, self.to_duckdb_relation(self.conn, **kwargs))

    def register_in_datafusion(
        self, ctx: datafusion.SessionContext, name: str | None = None, **kwargs
    ) -> None:
        """Register data in DataFusion.

        Args:
            ctx (datafusion.SessionContext): DataFusion session context instance.

        Returns:
            None
        """
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.ctx is None:
            if ctx is None:
                ctx = datafusion.SessionContext()
            self.ctx = ctx

        self.ctx.register_record_batches(
            name, [self.to_pyarrow_table(concat=True).to_batches()]
        )
        # return ctx

    def filter(
        self, filter_expr: str | pl.Expr | pa.compute.Expression
    ) -> (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | list[pl.DataFrame]
        | list[pl.LazyFrame]
        | list[pa.Table]
    ):
        """Filter data based on a filter expression.

        Args:
            filter_expr (str | pl.Expr | pa.compute.Expression): Filter expression.

        Returns:
            pl.DataFrame | pl.LazyFrame | pa.Table | list[pl.DataFrame] | list[pl.LazyFrame]
                | list[pa.Table]: Filtered data.
        """
        if isinstance(self._data, pl.DataFrame | pl.LazyFrame):
            pl_schema = (
                self._data.schema
                if isinstance(self._data, pl.DataFrame)
                else self._data.collect_schema()
            )
            filter_expr = (
                sql2polars_filter(filter_expr, pl_schema)
                if isinstance(filter_expr, str)
                else filter_expr
            )
            return self._data.filter(filter_expr)

        elif isinstance(self._data, pa.Table):
            pa_schema = self._data.schema
            filter_expr = (
                sql2pyarrow_filter(filter_expr, pa_schema)
                if isinstance(filter_expr, str)
                else filter_expr
            )
            return self._data.filter(filter_expr)

        if isinstance(self._data, str):
            if isinstance(self._data[0], pl.DataFrame | pl.LazyFrame):
                pl_schema = (
                    self._data.schema
                    if isinstance(self._data[0], pl.DataFrame)
                    else self._data[0].collect_schema()
                )
                filter_expr = (
                    sql2polars_filter(filter_expr, pl_schema)
                    if isinstance(filter_expr, str)
                    else filter_expr
                )
                return [d.filter(filter_expr) for d in self._data]
            elif isinstance(self._data[0], pa.Table):
                pa_schema = self._data[0].schema
                filter_expr = (
                    sql2pyarrow_filter(filter_expr, pa_schema)
                    if isinstance(filter_expr, str)
                    else filter_expr
                )
                return [d.filter(filter_expr) for d in self._data]


class BaseDatasetLoader(BaseFileLoader):
    """
    Base class for dataset loading operations supporting various file formats.
    This class provides a foundation for dataset loading operations across different file formats
    including CSV, Parquet, JSON, Arrow, and IPC.

    Args:
        path (str | list[str]): Path or list of paths to file(s).
        format (str, optional): File format extension (without dot).
        fs (AbstractFileSystem, optional): Filesystem instance for handling file operations.
        include_file_path (bool, optional): Include file path in the output DataFrame.
        concat (bool, optional): Concatenate multiple files into a single DataFrame.
        conn (duckdb.DuckDBPyConnection, optional): DuckDB connection instance.
        ctx (datafusion.SessionContext, optional): DataFusion session context instance.
        schema (pa.Schema, optional): PyArrow schema for the dataset.
        partitioning (str | list[str] | pds.Partitioning, optional): Dataset partitioning scheme.

        Examples:
        ```python
        dataset_loader = BaseDatasetLoader(
            path="s3://bucket/path/to/files",
            format="csv",
            include_file_path=True,
            concat=True,
            conn=duckdb.connect(),
            ctx=datafusion.SessionContext(),
            schema=pa.schema([
                pa.field("column1", pa.int64()),
                pa.field("column2", pa.string())
            ]),
            partitioning="hive"
        )
        data = dataset_loader.to_polars()
        ```
    Notes:
        - Supports multiple file formats including CSV, Parquet, JSON, Arrow, and IPC
        - Automatically handles filesystem initialization based on path protocol
        - Supports both single path and multiple path inputs
        - Supports loading data into DuckDB and DataFusion for SQL operations
        - Supports custom schema and partitioning for datasets

    """

    schema_: pa.Schema | None = None
    partitioning: str | list[str] | pds.Partitioning | None = None

    def to_pyarrow_dataset(
        self,
        **kwargs,
    ) -> pds.Dataset:
        """
        Convert data to PyArrow Dataset.

        Returns:
            pds.Dataset: PyArrow Dataset.
        """
        if self.format == ["csv", "arrow", "ipc"]:
            self._dataset = self.fs.pyarrow_dataset(
                self._path,
                format=self.format,
                schema=self.schema_,
                partitioning=self.partitioning,
                **kwargs,
            )
        elif self.format == "parquet":
            if self.fs.exists(posixpath.join(self._path, "_metadata")):
                self._dataset = self.fs.parquet_dataset(
                    self._path,
                    schema=self.schema_,
                    partitioning=self.partitioning,
                    **kwargs,
                )
            else:
                self._dataset = self.fs.pyarrow_dataset(
                    self._path,
                    format=self.format,
                    schema=self.schema_,
                    partitioning=self.partitioning,
                    **kwargs,
                )
        else:
            self._dataset = pds.dataset(
                self.to_pyarrow_table(**kwargs), schema=self.schema_
            )
        return self._dataset

    def to_pandas(self, **kwargs) -> pd.DataFrame:
        """
        Convert data to Pandas DataFrame.

        Returns:
            pd.DataFrame: Pandas DataFrame.
        """
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return self._dataset.to_table().to_pandas()

    def _to_polars_dataframe(self, **kwargs) -> pl.DataFrame:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return pl.from_arrow(self._dataset.to_table())

    def _to_polars_lazyframe(self, **kwargs) -> pl.LazyFrame:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return pl.scan_pyarrow_dataset(self._dataset)

    def to_polars(self, lazy: bool = True, **kwargs) -> pl.DataFrame | pl.LazyFrame:
        """
        Convert data to Polars DataFrame or LazyFrame.

        Args:
            lazy (bool, optional): Return a LazyFrame if True, else a DataFrame.

        Returns:
            pl.DataFrame | pl.LazyFrame: Polars DataFrame or LazyFrame.
        """
        return (
            self._to_polars_lazyframe(**kwargs)
            if lazy
            else self._to_polars_dataframe(**kwargs)
        )

    def to_pyarrow_table(self, **kwargs) -> pa.Table:
        """Convert data to PyArrow Table.

        Returns:
            pa.Table: PyArrow Table.
        """
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)
        return self._dataset.to_table()

    def to_pydala_dataset(self, **kwargs) -> ParquetDataset:
        if ParquetDataset is None:
            raise ImportError("pydala is not installed.")
        if not hasattr(self, "_pydala_dataset"):
            if not hasattr(self, "conn"):
                self.conn = duckdb.connect()
            self._pydala_dataset = self.fs.pydala_dataset(
                self._path,
                partitioning=self.partitioning,
                ddb_con=self.conn,
                **kwargs,
            )
        return self._pydala_dataset

    def to_duckdb_relation(
        self, conn: duckdb.DuckDBPyConnection | None = None, **kwargs
    ) -> duckdb.DuckDBPyRelation:
        """Convert data to DuckDB relation.

        Args:
            conn (duckdb.DuckDBPyConnection, optional): DuckDB connection instance.

        Returns:
            duckdb.DuckDBPyRelation: DuckDB relation.
        """
        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)

        return self.conn.from_arrow(self._dataset)

    def register_in_duckdb(
        self, conn: duckdb.DuckDBPyConnection, name: str | None = None, **kwargs
    ) -> None:
        """Register data in DuckDB.

        Args:
            conn (duckdb.DuckDBPyConnection): DuckDB connection instance.

        Returns:
            None
        """
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)

        self.conn.register(name, self._dataset)
        # return conn

    def register_in_datafusion(
        self, ctx: datafusion.SessionContext, name: str | None = None, **kwargs
    ) -> None:
        """Register data in DataFusion.

        Args:
            ctx (datafusion.SessionContext): DataFusion session context instance.

        Returns:
            None
        """
        if name is None:
            name = f"{self.format}:{self.path}"

        if self.ctx is None:
            if ctx is None:
                ctx = datafusion.SessionContext()
            self.ctx = ctx

        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset(**kwargs)

        self.ctx.register_dataset(name, self._dataset)
        # return self.tx

    def filter(self, filter_exp: str | pa.compute.Expression) -> pds.Dataset:
        if not hasattr(self, "_dataset"):
            self.to_pyarrow_dataset()
        if isinstance(filter_exp, str):
            filter_exp = sql2pyarrow_filter(filter_exp, self._dataset.schema)
        return self._dataset.filter(filter_exp)


class BaseFileWriter(BaseFileIO):
    data: (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pd.DataFrame
        | dict[str, Any]
        | list[pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict[str, Any]]
    ) | None = None
    basename: str | None = None
    concat: bool = False
    mode: str = "append"  # append, overwrite, delete_matching, error_if_exists

    def write(
        self,
        data: (
            pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pd.DataFrame
            | dict[str, Any]
            | list[
                pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict[str, Any]
            ]
        )
        | None = None,
        basename: str | None = None,
        concat: bool = False,
        mode: str = "append",
        **kwargs,
    ):
        self.fs.write_files(
            data=data if data is not None else self.data,
            path=self._path,
            basename=basename or self.basename,
            concat=concat or self.concat,
            mode=mode or self.mode,
            **kwargs,
        )


class BaseDatasetWriter(BaseFileWriter):
    data: (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | dict[str, Any]
        | list[
            pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pa.RecordBatch
            | pa.RecordBatchReader
            | pd.DataFrame
            | dict[str, Any]
        ]
    ) | None = None
    basename: str | None = None
    schema_: pa.Schema | None = None
    partition_by: str | list[str] | pds.Partitioning | None = None
    partitioning_flavor: str | None = None
    compression: str = "zstd"
    row_group_size: int | None = 250_000
    max_rows_per_file: int | None = 2_500_000
    concat: bool = False
    mode: str = "append"  # append, overwrite, delete_matching, error_if_exists
    is_pydala_dataset: bool = False

    def write(
        self,
        data: (
            pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pa.RecordBatch
            | pa.RecordBatchReader
            | pd.DataFrame
            | dict[str, Any]
            | list[
                pl.DataFrame
                | pl.LazyFrame
                | pa.Table
                | pa.RecordBatch
                | pa.RecordBatchReader
                | pd.DataFrame
                | dict[str, Any]
            ]
        )
        | None = None,
        unique: bool | list[str] | str = False,
        delta_subset: str | None = None,
        alter_schema: bool = False,
        update_metadata: bool = True,
        timestamp_column: str | None = None,
        verbose: bool = False,
        **kwargs,
    ):
        """
        Write data to dataset.

        Args:
            data (pl.DataFrame | pl.LazyFrame | pa.Table | pa.RecordBatch | pa.RecordBatchReader | pd.DataFrame |
                dict[str, Any] | list[pl.DataFrame | pl.LazyFrame | pa.Table | pa.RecordBatch | pa.RecordBatchReader |
                pd.DataFrame | dict[str, Any]] | None, optional): Data to write.
            unique (bool | list[str] | str, optional): Unique columns for deduplication.
            delta_subset (str | None, optional): Delta subset for incremental updates.
            alter_schema (bool, optional): Alter schema for compatibility.
            update_metadata (bool, optional): Update metadata.
            timestamp_column (str | None, optional): Timestamp column for updates.
            verbose (bool, optional): Verbose output.
            **kwargs: Additional keyword arguments.
        """
        basename = kwargs.pop("basename", self.basename)
        schema = kwargs.pop("schema", self.schema_)
        partition_by = kwargs.pop("partition_by", self.partition_by)
        partitioning_flavor = kwargs.pop(
            "partitioning_flavor", self.partitioning_flavor
        )
        compression = kwargs.pop("compression", self.compression)
        row_group_size = kwargs.pop("row_group_size", self.row_group_size)
        max_rows_per_file = kwargs.pop("max_rows_per_file", self.max_rows_per_file)
        concat = kwargs.pop("concat", self.concat)
        mode = kwargs.pop("mode", self.mode)

        if not self.is_pydala_dataset:
            self.fs.write_pyarrow_dataset(
                data=data if data is not None else self.data,
                path=self._path,
                basename=basename or self.basename,
                schema=schema or self.schema_,
                partition_by=partition_by or self.partition_by,
                partitioning_flavor=partitioning_flavor or self.partitioning_flavor,
                format=self.format,
                compression=compression or self.compression,
                row_group_size=row_group_size or self.row_group_size,
                max_rows_per_file=max_rows_per_file or self.max_rows_per_file,
                concat=concat or self.concat,
                mode=mode or self.mode,
                **kwargs,
            )
        else:
            self.fs.write_pydala_dataset(
                data=data if data is not None else self.data,
                path=self._path,
                mode=mode,
                basename=basename or self.basename,
                schema=schema or self.schema_,
                partition_flavor=partitioning_flavor or self.partitioning_flavor,
                partition_by=partition_by or self.partition_by,
                compression=compression or self.compression,
                row_group_size=row_group_size or self.row_group_size,
                max_rows_per_file=max_rows_per_file or self.max_rows_per_file,
                concat=concat or self.concat,
                unique=unique,
                delta_subset=delta_subset,
                alter_schema=alter_schema,
                update_metadata=update_metadata,
                timestamp_column=timestamp_column,
                verbose=verbose,
                **kwargs,
            )



ALL_DATA_STORES = [
    "sqlalchemy",
    "sqlite",
    "postgresql",
    "mysql",
    "mongodb",
    "local",
    "memory",
]


class DataStore:
    def __init__(self, type: str | None = None, engine_or_uri: str | None = None):
        self.type = type or "memory"
        self.engine_or_uri = engine_or_uri
        self.sqla_engine = None

        if self.type not in ALL_DATA_STORES:
            raise ValueError(
                f"Invalid data store type: {type}. Valid data store types are: {ALL_DATA_STORES}"
            )
        if (
            type in ["sqlalchemy", "sqlite", "postgresql", "mysql", "mongodb"]
            and not engine_or_uri
        ):
            raise ValueError(f"Data store type {type} requires an engine or uri")

    def _setup_sqlalchemy(self):
        from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore
        from sqlalchemy.ext.asyncio import create_async_engine

        self.sqla_engine = create_async_engine(self.engine_or_uri)
        self._data_store = SQLAlchemyDataStore(self.sqla_engine)

    def _setup_mongodb(self, uri: str):
        from apscheduler.datastores.mongodb import MongoDBDataStore

        self._data_store = MongoDBDataStore(self.engine_or_uri)

    def _setup_memory(self):
        from apscheduler.datastores.memory import MemoryDataStore

        self._data_store = MemoryDataStore()

    def setup(
        self,
    ):
        if self.type in ["sqlalchemy", "sqlite", "postgresql", "mysql"]:
            self._setup_sqlalchemy()
        elif self.type == "mongodb":
            self._setup_mongodb()
        else:
            self._setup_memory()

    def get(self) -> tuple:
        return self._data_store, self.sqla_engine


def setup_data_store(type: str, engine_or_uri: str) -> tuple:
    ds = DataStore(type=type, engine_or_uri=engine_or_uri)
    ds.setup()
    return ds.get()



# import datetime as dt


import pyarrow as pa
import pyarrow.dataset as pds
from deltalake import DeltaTable

# from hamilton.function_modifiers import dataloader

from ..base import BaseDatasetLoader

# from ..utils import get_dataframe_metadata, get_delta_metadata


class DeltaTableLoader(BaseDatasetLoader):
    """Delta table loader.

    This class is responsible for loading Delta tables into several dataframe formats,
    duckdb and datafusion.

    """

    delta_table: DeltaTable | None = None

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "delta"

        self.delta_table = DeltaTable(
            self._raw_path,
            storage_options=self.storage_options.to_object_store_kwargs(),
        )

    @property
    def dt(self) -> DeltaTable:
        return self.delta_table

    def to_pyarrow_dataset(self) -> pds.Dataset:
        """Converts the DeltaTable to a PyArrow Dataset."""
        return self.delta_table.to_pyarrow_dataset()

    def to_pyarrow_table(self) -> pa.Table:
        """Converts the DeltaTable to a PyArrow Table."""
        return self.delta_table.to_pyarrow_table()

    # def to_polars(self, lazy: bool = True) -> pl.DataFrame | pl.LazyFrame:
    #     """Converts the DeltaTable to a Polars DataFrame."""
    #     if lazy:
    #         return pl.scan_pyarrow_dataset(self.to_pyarrow_dataset())

    #     return pl.from_arrow(self.to_pyarrow_table())

    # def to_pandas(self) -> pd.DataFrame:
    #     """Converts the DeltaTable to a Pandas DataFrame."""
    #     return self.to_pyrrow_table().to_pandas()

    # def to_duckdb_relation(
    #     self, conn: duckdb.DuckDBPyConnection | None = None, lazy: bool = True
    # ) -> duckdb.DuckDBPyRelation:
    #     """Converts the DeltaTable to a DuckDB relation."""
    #     if conn is None:
    #         conn = duckdb.connect()
    #     if lazy:
    #         return conn.from_arrow(self.to_pyarrow_dataset())
    #     return conn.from_arrow(self.to_pyarrow_table())

    # def register_in_duckdb(
    #     self,
    #     conn: duckdb.DuckDBPyConnection | None = None,
    #     name: str | None = None,
    #     lazy: bool = True,
    # ) -> duckdb.DuckDBPyConnection:
    #     """Registers the DeltaTable in a DuckDB connection."""
    #     if name is None:
    #         name = f"{self.format}:{self.path}"
    #     if conn is None:
    #         conn = duckdb.connect()
    #     table = self.to_duckdb_relation(conn=conn, lazy=lazy)
    #     conn.register(name, table)
    #     return conn

    # def register_in_datafusion(
    #     self,
    #     ctx: dtf.SessionContext | None = None,
    #     name: str | None = None,
    #     lazy: bool = True,
    # ) -> dtf.SessionContext:
    #     """Registers the DeltaTable in a DataFusion context."""
    #     if name is None:
    #         name = f"{self.format}:{self.path}"
    #     if ctx is None:
    #         ctx = dtf.SessionContext()
    #     if lazy:
    #         ctx.register_dataset(name, self.to_pyarrow_dataset())
    #     else:
    #         ctx.register_record_batches(name, [self.to_pyarrow_table().to_batches()])
    #         # table = table.to_table()
    #     # ctx.register_dataset(name, table)
    #     return ctx



from munch import Munch, munchify
from pydantic import Field

from ..base import BaseConfig


class PipelineTrackerConfig(BaseConfig):
    project_id: int | None = None
    # version: str | None = None
    dag_name: str | None = None
    tags: dict | Munch = Field(default_factory=dict)

    def model_post_init(self, __context):
        self.tags = munchify(self.tags)



from pydantic import BaseModel, Field
import datetime as dt


class PipelineRun(BaseModel):
    inputs: dict | None = None
    final_vars: list | None = None
    executor: str | None = None
    config: dict | None = None
    with_tracker: bool | None = None
    with_opentelemetry: bool | None = None


class PipelineAddJob(PipelineRun):
    result_expiration_time: float | dt.timedelta = 0


class PipelineSchedule(PipelineRun):
    trigger_type: str | None = None
    paused: bool = False
    coalesce: str = "latest"
    misfire_grace_time: float | dt.timedelta | None = None
    max_jitter: float | dt.timedelta | None = Field
    max_running_jobs: int | None = None
    conflict_policy: str = "do_nothing"


class PipelineManagerNew(BaseModel):
    overwrite: bool = False


class PipelineManagerImportExport(BaseModel):
    path: str
    name: str | None = None
    names: list[str] | None = None
    cfg_dir: str = "conf"
    pipeline_dir: str = "pipelines"
    storage_options: dict | None = None
    overwrite: bool = False


class PipelineDelete(BaseModel):
    name: str
    cfg: bool = False
    module: bool = False


class PipelineManagerSummary(BaseModel):
    cfg: bool = True
    module: bool = True



import dill
from sanic import Blueprint, SanicException
from sanic_ext import validate


from sanic.response import json, raw, html

from ...pipeline import Pipeline, PipelineManager
from ..models.pipeline import (
    PipelineRun,
    PipelineAddJob,
    PipelineSchedule,
    PipelineManagerNew,
    PipelineManagerImportExport,
    PipelineDelete,
    PipelineManagerSummary,
)

bp = Blueprint("api_flowerpower_pipeline", url_prefix="api/pipeline")


@bp.post("run/<name>")
@validate(json=PipelineRun)
async def run(
    request,
    name: str,
    body: PipelineRun,
):
    try:
        with Pipeline(
            name=name,
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
        ) as pipeline:
            final_vars = pipeline.run(**body.model_dump())

        final_vars = dill.dumps(final_vars)
        # {k: dill.dumps(v) for k, v in final_vars.items()}
        return raw(final_vars)
        # return json({"status": "success", "message": "Pipeline ran successfully"})
    except Exception as e:
        raise SanicException(str(e))


@bp.post("run-job/<name>")
@validate(json=PipelineRun)
async def run_job(
    request,
    name: str,
    body: PipelineRun,
):
    try:
        with Pipeline(
            name=name,
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
        ) as pipeline:
            final_vars = pipeline.run_job(**body.model_dump())
        final_vars = dill.dumps(final_vars)
        return raw(final_vars)
    except Exception as e:
        raise SanicException(str(e))


@bp.post("/add-job/<name>")
@validate(json=PipelineAddJob)
async def add_job(
    request,
    name: str,
    body: PipelineAddJob,
):
    try:
        with Pipeline(
            name=name,
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as pipeline:
            id_ = pipeline.add_job(**body.model_dump())
        return json({"job_id": str(id_)})
    except Exception as e:
        raise SanicException(str(e))


@bp.post("/schedule/<name>")
@validate(json=PipelineSchedule)
async def schedule(
    request,
    name: str,
    body: PipelineSchedule,
):

    try:
        with Pipeline(
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as manager:
            id_ = manager.schedule(name, **body.model_dump())
        return json({"schedule_id": str(id_)})
    except Exception as e:
        raise SanicException(str(e))


@bp.patch("/schedule/<name>")
@validate(json=PipelineSchedule)
async def update_schedule(
    request,
    name: str,
    body: PipelineSchedule,
):
    params = body.model_dump()
    params.update({"overwrite": True})

    try:
        with Pipeline(
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as manager:
            id_ = manager.schedule(name, **params)
        return json({"schedule_id": str(id_)})
    except Exception as e:
        raise SanicException(str(e))


@bp.post("new/<name>")
@validate(json=PipelineManagerNew)
async def new(
    request,
    name: str,
    body: PipelineManagerNew,
):

    try:
        with PipelineManager(
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as manager:
            manager.new(name, **body.model_dump())
        return json({"status": f"Pipeline {name} created"})
    except Exception as e:
        raise SanicException(str(e))


@bp.post("/import")
@validate(json=PipelineManagerImportExport)
async def import_pipeline(
    request,
    body: PipelineManagerImportExport,
):
    params = body.model_dump()
    name = params.pop("name", None)
    names = params.pop("names", None)
    path = params.pop("path", None)

    if isinstance(names, str):
        names = names.split(",")
    try:

        with PipelineManager(
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as manager:
            if name:
                manager.import_pipeline(name=name, path=path, **params)
                return json({"status": f"Pipeline {name} imported from {path}"})
            elif names:
                manager.import_many(names, path=path, **params)
                return json(
                    {"status": f"Pipelines {', '.join(names)} imported from {path}"}
                )
            else:
                manager.import_all(path=path, **params)
                return json({"status": f"All pipelines imported from {path}"})

    except Exception as e:
        raise SanicException(str(e))


@bp.post("/export")
@validate(json=PipelineManagerImportExport)
async def export_pipeline(
    request,
    body: PipelineManagerImportExport,
):
    params = body.model_dump()
    path = params.pop("path", None)
    name = params.pop("name", None)
    names = params.pop("names", None)

    if isinstance(names, str):
        names = names.split(",")
    try:
        with PipelineManager(
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as manager:
            if name:
                manager.export_pipeline(name, path=path, **params)
                return json({"status": f"Pipeline {name} exported to {path}"})
            elif names:
                manager.export_many(names, path=path, **params)
                return json(
                    {"status": f"Pipelines {', '.join(names)} exported to {path}"}
                )
            else:
                manager.export_all(path=path, **params)
                return json({"status": f"All pipelines exported to {path}"})
    except Exception as e:
        raise SanicException(str(e))


@bp.delete("/delete/<name>")
@validate(json=PipelineDelete)
async def delete(
    request,
    name: str,
    body: PipelineDelete,
):
    try:
        with Pipeline(
            name=name,
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as pipeline:
            pipeline.delete(**body.model_dump())
        return json({"status": f"Pipeline {name} deleted"})
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/summary/<name>")
@validate(query=PipelineManagerSummary)
async def summary(
    request,
    query: PipelineManagerSummary,
    name: str | None = None,
):

    as_ = request.args.get("as", None)
    if as_ == "html":
        to_html = True
        to_svg = False
    if as_ == "svg":
        to_svg = True
        to_html = False
    else:
        to_html = False
        to_svg = False

    try:
        if name == "all":
            with PipelineManager(
                base_dir=request.app.config.BASE_DIR,
                storage_options=request.app.config.STORAGE_OPTIONS,
                cfg_dir=request.app.config.CFG_DIR,
                pipelines_dir=request.app.config.PIPELINES_DIR,
            ) as manager:
                if to_html or to_svg:
                    summary = manager.show_summary(
                        to_html=to_html, to_svg=to_svg, **query.model_dump()
                    )
                summary = manager.get_summary(**query.model_dump())

        else:
            with Pipeline(
                name=name,
                base_dir=request.app.config.BASE_DIR,
                storage_options=request.app.config.STORAGE_OPTIONS,
                cfg_dir=request.app.config.CFG_DIR,
                pipelines_dir=request.app.config.PIPELINES_DIR,
            ) as pipeline:

                summary = pipeline.get_summary(
                    to_html=to_html, to_svg=to_svg, **query.model_dump()
                )
        if to_html:
            return html(summary)
        elif to_svg:
            return raw(summary)
        return json(summary)
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/show/<name>")
async def show(
    request,
    name: str,
):

    format = request.args.get("format", "png")

    try:
        with Pipeline(
            name=name,
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as pipeline:
            pipeline_dag = pipeline.show_dag(raw=True)
        return raw(pipeline_dag.pipe(format))
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/pipelines")
async def pipelines(
    request,
):
    show = request.args.get("show", False)
    format_ = request.args.get("format", "html")
    to_html = format_.lower() == "html"
    to_svg = format_.lower() == "svg"
    try:
        with PipelineManager(
            base_dir=request.app.config.BASE_DIR,
            storage_options=request.app.config.STORAGE_OPTIONS,
            cfg_dir=request.app.config.CFG_DIR,
            pipelines_dir=request.app.config.PIPELINES_DIR,
        ) as manager:
            if show:
                pipelines = manager._all_pipelines(
                    show=True, to_html=to_html, to_svg=to_svg
                )
            else:
                pipelines = manager.list_pipelines()
        if to_html:
            return html(pipelines)
        elif to_svg:
            return raw(pipelines)
        return json(pipelines)
    except Exception as e:
        raise SanicException(str(e))



import datetime as dt

# from tzlocal import get_localzone

ALL_TRIGGER_TYPES = [
    "cron",
    "interval",
    "calendarinterval",
    "date",
]
ALL_TRIGGER_KWARGS = {
    "cron": [
        "crontab",
        "year",
        "month",
        "week",
        "day",
        "day_of_week",
        "hour",
        "minute",
        "second",
        "start_time",
        "end_time",
        "timezone",
    ],
    "interval": [
        "weeks",
        "days",
        "hours",
        "minutes",
        "seconds",
        "microseconds",
        "start_time",
        "end_time",
    ],
    "calendarinterval": [
        "years",
        "months",
        "weeks",
        "days",
        "hour",
        "minute",
        "second",
        "start_date",
        "end_date",
        "timezone",
    ],
    "date": [
        "run_time",
    ],
}


class Trigger:
    def __init__(
        self,
        type_: str,
    ):
        if type_ not in ALL_TRIGGER_TYPES:
            raise ValueError(
                f"Invalid trigger type: {type_}. Valid trigger types are: {ALL_TRIGGER_TYPES}"
            )
        self.trigger_type = type_

    def _check_kwargs(self, **kwargs):
        for k, v in kwargs.items():
            if k not in ALL_TRIGGER_KWARGS[self.trigger_type]:
                raise ValueError(
                    f"Invalid argument: {k}. Valid arguments are: {ALL_TRIGGER_KWARGS[self.trigger_type]}"
                )

    def _filter_kwargs(self, **kwargs):
        return {
            k: v
            for k, v in kwargs.items()
            if k in ALL_TRIGGER_KWARGS[self.trigger_type] and v is not None
        }

    def get(self, **kwargs):
        #
        kwargs = self._filter_kwargs(**kwargs)
        self._check_kwargs(**kwargs)

        if self.trigger_type == "cron":
            return self._get_cron_trigger(**kwargs)
        elif self.trigger_type == "interval":
            return self._get_interval_trigger(**kwargs)
        elif self.trigger_type == "calendarinterval":
            return self._get_calendar_trigger(**kwargs)
        elif self.trigger_type == "date":
            return self._get_date_trigger(**kwargs)

    def _get_cron_trigger(
        self,
        **kwargs,
    ):
        from apscheduler.triggers.cron import CronTrigger

        crontab = kwargs.pop("crontab", None)
        print(crontab)

        if crontab is not None:
            return CronTrigger.from_crontab(crontab)  # , kwargs)
        else:
            return CronTrigger(
                **kwargs,
            )

    def _get_interval_trigger(
        self,
        **kwargs,
    ):
        from apscheduler.triggers.interval import IntervalTrigger

        return IntervalTrigger(
            **kwargs,
        )

    def _get_calendar_trigger(
        self,
        **kwargs,
    ):
        from apscheduler.triggers.calendarinterval import CalendarIntervalTrigger

        return CalendarIntervalTrigger(
            **kwargs,
        )

    def _get_date_trigger(self, **kwargs):
        from apscheduler.triggers.date import DateTrigger

        if "run_time" not in kwargs:
            kwargs["run_time"] = dt.datetime.now()
        return DateTrigger(**kwargs)


def get_trigger(type_: str, **kwargs):
    trigger = Trigger(type_)
    return trigger.get(**kwargs)



import configparser
import os

import yaml
from fsspec import AbstractFileSystem, filesystem
from pydantic import BaseModel


class BaseStorageOptions(BaseModel):
    protocol: str

    def to_dict(self) -> dict:
        return {k: v for k, v in self.model_dump().items() if v is not None}

    @classmethod
    def from_yaml(
        cls, path: str, fs: AbstractFileSystem = None
    ) -> "BaseStorageOptions":
        with fs.open(path, "r") as f:
            data = yaml.safe_load(f)
        return cls(**data)

    def to_yaml(self, path: str, fs: AbstractFileSystem = None):
        if fs is None:
            fs = filesystem("file")
        data = self.to_dict()
        with fs.open(path, "w") as f:
            yaml.safe_dump(data, f)

    def to_filesystem(self) -> AbstractFileSystem:
        return filesystem(**self.to_dict())

    def update(self, **kwargs):
        self = self.model_copy(update=kwargs)


class AzureStorageOptions(BaseStorageOptions):
    pass


class GcsStorageOptions(BaseStorageOptions):
    pass


class AwsStorageOptions(BaseStorageOptions):
    protocol: str = "s3"
    access_key_id: str | None = None
    secret_access_key: str | None = None
    session_token: str | None = None
    endpoint_url: str | None = None
    region: str | None = None
    allow_invalid_certificates: bool | None = None
    allow_http: bool | None = None
    profile: str | None = None

    def model_post_init(self, __context):
        if self.profile is not None:
            super().__init__(
                **self.from_aws_credentials(
                    profile=self.profile,
                    allow_invalid_certificates=self.allow_invalid_certificates,
                    allow_http=self.allow_http,
                ).to_dict()
            )

    @classmethod
    def from_aws_credentials(
        cls,
        profile: str,
        allow_invalid_certificates: bool = False,
        allow_http: bool = False,
    ) -> "AwsStorageOptions":
        cp = configparser.ConfigParser()
        cp.read(os.path.expanduser("~/.aws/credentials"))
        cp.read(os.path.expanduser("~/.aws/config"))
        if profile not in cp:
            raise ValueError(f"Profile '{profile}' not found in AWS credentials file")

        return cls(
            protocol="s3",
            access_key_id=cp[profile].get("aws_access_key_id", None),
            secret_access_key=cp[profile].get("aws_secret_access_key", None),
            session_token=cp[profile].get("aws_session_token", None),
            endpoint_url=cp[profile].get("aws_endpoint_url", None)
            or cp[profile].get("endpoint_url", None)
            or cp[profile].get("aws_endpoint", None)
            or cp[profile].get("endpoint", None),
            region=(
                cp[profile].get("region", None)
                or cp[f"profile {profile}"].get("region", None)
                if f"profile {profile}" in cp
                else None
            ),
            allow_invalid_certificates=allow_invalid_certificates,
            allow_http=allow_http,
        )

    @classmethod
    def from_env(cls) -> "AwsStorageOptions":
        return cls(
            access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
            secret_access_key=os.getenv("AWS_SECRET_ACCESS_KEY"),
            session_token=os.getenv("AWS_SESSION_TOKEN"),
            endpoint_url=os.getenv("AWS_ENDPOINT_URL"),
            region=os.getenv("AWS_DEFAULT_REGION"),
            allow_invalid_certificates="true"
            == (os.getenv("ALLOW_INVALID_CERTIFICATES", "False").lower()),
            allow_http="true" == (os.getenv("AWS_ALLOW_HTTP", "False").lower()),
        )

    def to_fsspec_kwargs(self) -> dict:
        fsspec_kwargs = {
            "key": self.access_key_id,
            "secret": self.secret_access_key,
            "token": self.session_token,
            "endpoint_url": self.endpoint_url,
            "client_kwargs": {
                "region_name": self.region,
                "verify": (
                    not self.allow_invalid_certificates
                    if self.allow_invalid_certificates is not None
                    else False
                ),
                "use_ssl": not self.allow_http if self.allow_http is not None else True,
            },
        }
        return {k: v for k, v in fsspec_kwargs.items() if v is not None}

    def to_object_store_kwargs(self, conditional_put="etag") -> dict:
        object_store_kwargs = {
            k: str(v)
            for k, v in self.to_dict().items()
            if v is not None and k != "protocol"
        }
        if len(conditional_put):
            object_store_kwargs["conditional_put"] = conditional_put
        return object_store_kwargs

    def to_env(self) -> None:
        env = {
            "AWS_ACCESS_KEY_ID": self.access_key_id,
            "AWS_SECRET_ACCESS_KEY": self.secret_access_key,
            "AWS_SESSION_TOKEN": self.session_token,
            "AWS_ENDPOINT_URL": self.endpoint_url,
            "AWS_DEFAULT_REGION": self.region,
            "ALLOW_INVALID_CERTIFICATES": str(self.allow_invalid_certificates),
            "AWS_ALLOW_HTTP": str(self.allow_http),
        }
        env = {k: v for k, v in env.items() if v is not None}
        os.environ.update(env)

    def to_filesystem(self):
        return filesystem(self.protocol, **self.to_fsspec_kwargs())


class GitHubStorageOptions(BaseStorageOptions):
    protocol: str = "github"
    org: str | None = None
    repo: str | None = None
    sha: str | None = None

    @classmethod
    def from_env(cls) -> "GitHubStorageOptions":
        return cls(
            protocol="github",
            org=os.getenv("GITHUB_ORG"),
            repo=os.getenv("GITHUB_REPO"),
            sha=os.getenv("GITHUB_SHA"),
        )

    def to_env(self) -> None:
        os.environ.update(
            {"GITHUB_ORG": self.org, "GITHUB_REPO": self.repo, "GITHUB_SHA": self.sha}
        )


class GitLabStorageOptions(BaseStorageOptions):
    protocol: str = "gitlab"
    base_url: str = "https://gitlab.com"
    access_token: str | None = None
    project_id: str | int | None = None
    project_name: str | None = None

    @classmethod
    def from_env(cls) -> "GitLabStorageOptions":
        return cls(
            protocol="gitlab",
            base_url=os.getenv("GITLAB_BASE_URL"),
            access_token=os.getenv("GITLAB_ACCESS_TOKEN"),
            project_id=os.getenv("GITLAB_PROJECT_ID"),
            project_name=os.getenv("GITLAB_PROJECT_NAME"),
        )

    def model_post_init(self, __context):
        if self.project_id is None and self.project_name is None:
            raise ValueError("Either 'project_id' or 'project_name' must be provided")


class LocalStorageOptions(BaseStorageOptions):
    protocol: str = "file"


def from_dict(
    protocol: str, storage_options: dict
) -> (
    AwsStorageOptions
    | AzureStorageOptions
    | GcsStorageOptions
    | GitHubStorageOptions
    | GitLabStorageOptions
    | LocalStorageOptions
):
    if protocol == "s3":
        return AwsStorageOptions(**storage_options)
    elif protocol == "az" or protocol == "abfs" or protocol == "adl":
        return AzureStorageOptions(**storage_options)
    elif protocol == "gs" or protocol == "gcs":
        return GcsStorageOptions(**storage_options)
    elif protocol == "github":
        return GitHubStorageOptions(**storage_options)
    elif protocol == "gitlab":
        return GitLabStorageOptions(**storage_options)
    elif protocol == "file":
        return LocalStorageOptions()
    else:
        raise ValueError(f"Unsupported protocol: {protocol}")


def from_env(
    protocol: str,
) -> (
    AwsStorageOptions
    | AzureStorageOptions
    | GcsStorageOptions
    | GitHubStorageOptions
    | GitLabStorageOptions
    | LocalStorageOptions
):
    if protocol == "s3":
        return AwsStorageOptions.from_env()
    elif protocol == "github":
        return GitHubStorageOptions.from_env()
    elif protocol == "gitlab":
        return GitLabStorageOptions.from_env()
    elif protocol == "file":
        return LocalStorageOptions()
    else:
        raise ValueError(f"Unsupported protocol: {protocol}")


class StorageOptions(BaseModel):
    storage_options: BaseStorageOptions

    def __init__(self, **data):
        protocol = data.get("protocol")
        if protocol is None and "storage_options" not in data:
            raise ValueError("protocol must be specified")

        if "storage_options" not in data:
            if protocol == "s3":
                storage_options = AwsStorageOptions(**data)
            elif protocol == "github":
                storage_options = GitHubStorageOptions(**data)
            elif protocol == "gitlab":
                storage_options = GitLabStorageOptions(**data)
            elif protocol in ["az", "abfs", "adl"]:
                storage_options = AzureStorageOptions(**data)
            elif protocol in ["gs", "gcs"]:
                storage_options = GcsStorageOptions(**data)
            elif protocol == "file":
                storage_options = LocalStorageOptions(**data)
            else:
                raise ValueError(f"Unsupported protocol: {protocol}")

            super().__init__(storage_options=storage_options)
        else:
            super().__init__(**data)

    @classmethod
    def from_yaml(cls, path: str, fs: AbstractFileSystem = None) -> "StorageOptions":
        with fs.open(path, "r") as f:
            data = yaml.safe_load(f)
        return cls(**data)

    @classmethod
    def from_env(cls, protocol: str) -> "StorageOptions":
        if protocol == "s3":
            return cls(storage_options=AwsStorageOptions.from_env())
        elif protocol == "github":
            return cls(storage_options=GitHubStorageOptions.from_env())
        elif protocol == "gitlab":
            return cls(storage_options=GitLabStorageOptions.from_env())
        elif protocol == "file":
            return cls(storage_options=LocalStorageOptions())
        else:
            raise ValueError(f"Unsupported protocol: {protocol}")

    def to_filesystem(self) -> AbstractFileSystem:
        return self.storage_options.to_filesystem()

    def to_dict(self) -> dict:
        return self.storage_options.to_dict()

    def to_object_store_kwargs(self, conditional_put="etag") -> dict:
        return self.storage_options.to_object_store_kwargs(
            conditional_put=conditional_put
        )



from ..base import BaseDatasetLoader, BaseFileLoader


class ParquetFileLoader(BaseFileLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "parquet"

    def __call__(self, **kwargs):
        return self.to_pyarrow_table(**kwargs)


class ParquetDatasetLoader(BaseDatasetLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "parquet"

    def __call__(self, **kwargs):
        return self.to_pyarrow_dataset(**kwargs)



import duckdb
import pandas as pd
import polars as pl
import pyarrow as pa
from pydantic import BaseModel
import os
from typing import Literal
from ...utils.misc import _dict_to_dataframe


class DuckDBSaver(BaseModel):
    path: str | None = None
    table_name: str
    mode: Literal["overwrite", "append", "fail", "merge", "update"] = "append"
    key_columns: list[str] | None = None  # Required for merge/update modes
    conn: duckdb.DuckDBPyConnection | None = None
    host: str | None = None
    port: int | None = None
    socket: str | None = None
    user: str | None = None
    password: str | None = None
    database: str | None = None

    def _is_sqlite_file(self, path: str) -> bool:
        """Check if file is SQLite by reading first 16 bytes for SQLite header."""
        if not os.path.exists(path):
            return False
        with open(path, "rb") as f:
            header = f.read(16)
        return header.startswith(b"SQLite format 3\x00")

    def model_post_init(self, __context):
        if self.path is None:
            self.path = ":memory:"
            self.conn = duckdb.connect(self.path)
        else:
            if os.path.exists(self.path):
                if self._is_sqlite_file(self.path):
                    # For SQLite files, create in-memory DuckDB and attach SQLite
                    self.conn = duckdb.connect(":memory:")
                    self.conn.execute("INSTALL sqlite; LOAD sqlite;")
                    self.conn.execute(
                        f"ATTACH '{self.path}' AS sqlite_db (TYPE SQLITE);"
                    )
                else:
                    # For DuckDB files, connect directly
                    self.conn = duckdb.connect(self.path)
            else:
                # For new files, create new DuckDB database
                self.conn = duckdb.connect(self.path)

        if self.host is not None:
            self._update_from_env()

        if self.host is not None:
            if self.port == 5432:
                self._attach_postgres()
            elif self.port == 3306:
                self._attach_mysql()

    def _update_from_env(self):
        self.host = (
            os.getenv("POSTGRES_HOST") or os.getenv("PGHOST") or os.getenv("MYSQL_HOST")
        )
        self.port = (
            os.getenv("POSTGRES_PORT")
            or os.getenv("PGPORT")
            or os.getenv("MYSQL_TCP_PORT")
        )
        self.user = (
            os.getenv("POSTGRES_USER") or os.getenv("PGUSER") or os.getenv("MYSQL_USER")
        )
        self.password = (
            os.getenv("POSTGRES_PASSWORD")
            or os.getenv("PGPASSWORD")
            or os.getenv("MYSQL_PW")
        )
        self.database = (
            os.getenv("POSTGRES_DB")
            or os.getenv("PGDATABASE")
            or os.getenv("MYSQL_DATABASE")
        )
        self.socket = os.getenv("MYSQL_UNIX_PORT")

    def _attach_postgres(self):
        self.conn.execute("INSTALL postgres; LOAD postgres;")
        connection_string = " ".join(
            f"{k}={v}"
            for k, v in {
                "host": self.host,
                "port": self.port,
                "user": self.user,
                "password": self.password,
                "database": self.database,
            }.items()
            if v is not None
        )
        self.conn.execute(f"ATTACH {connection_string} AS postgres_db;")

    def _attach_mysql(self):
        self.conn.execute("INSTALL mysql; LOAD mysql;")
        connection_string = " ".join(
            f"{k}={v}"
            for k, v in {
                "host": self.host,
                "port": self.port,
                "socket": self.socket,
                "user": self.user,
                "password": self.password,
                "database": self.database,
            }.items()
            if v is not None
        )
        self.conn.execute(f"ATTACH {connection_string} AS mysql_db;")

    def _validate_merge_params(self):
        if self.mode in ("merge", "update") and not self.key_columns:
            # raise ValueError(f"key_columns must be specified for mode '{self.mode}'")
            # key columns should be all columns in the table
            self.key_columns = self.conn.table(self.table_name).columns

    def _table_exists(self) -> bool:
        tables_query = (
            "SELECT name FROM sqlite_master"
            if self._is_sqlite_file(self.path)
            else "SELECT table_name as name FROM information_schema.tables"
        )
        result = self.conn.execute(
            f"{tables_query} WHERE name='{self.table_name}'"
        ).fetchone()
        return bool(result)

    def _handle_existing_table(self) -> bool:
        """Handle existing table based on mode. Returns True if table exists."""
        exists = self._table_exists()

        if exists:
            if self.mode == "fail":
                raise ValueError(f"Table {self.table_name} already exists")
            elif self.mode == "overwrite":
                self.conn.execute(f"DROP TABLE {self.table_name}")
                return False
            elif self.mode in ("append", "merge", "update"):
                return True
        return False

    def write(
        self,
        data: (
            pd.DataFrame
            | pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pa.RecordBatch
            | pa.RecordBatchReader
            | dict
            | list[
                pd.DataFrame
                | pl.DataFrame
                | pl.LazyFrame
                | pa.Table
                | pa.RecordBatch
                | pa.RecordBatchReader
                | dict
            ]
        ),
        mode: Literal["overwrite", "append", "fail", "merge", "update"] = None,
        key_columns: list[str] | None = None,
    ):
        """Write data to database with specified merge strategy.

        Args:
            data: Data to write. Supports pandas/polars DataFrames, pyarrow Tables/Batches, or dicts
        """
        mode = mode or self.mode
        key_columns = key_columns or self.key_columns

        if isinstance(data, dict):
            data = _dict_to_dataframe(data)

        if not isinstance(data, list):
            data = [data]
        if isinstance(data[0], dict):
            data = [_dict_to_dataframe(d) for d in data]
        if isinstance(data[0], pa.LazyFrame):
            data = [d.collect() for d in data]
        if isinstance(data[0], pl.DataFrame):
            data = pl.concat(data, how="diagonal_relaxed").to_arrow()
        if isinstance(data[0], pd.DataFrame):
            data = pa.concat_tables(
                [pa.Table.from_pandas(d, preserve_index=False) for d in data],
                promote_options="permissive",
            )
        if isinstance(data[0], pa.RecordBatch | pa.RecordBatchReader):
            data = pa.Table.from_batches(data)
        if isinstance(data[0], pa.Table):
            data = pa.concat_tables(data, promote_options="permissive")

        self._validate_merge_params()
        table_exists = self._handle_existing_table()

        # Convert input to temporary table
        temp_name = f"temp_{self.table_name}"
        self.conn.execute(f"DROP TABLE IF EXISTS {temp_name}")
        self.conn.execute(f"CREATE TEMP TABLE {temp_name} AS SELECT * FROM data")

        if not table_exists:
            # Simple case - just rename temp table
            self.conn.execute(f"ALTER TABLE {temp_name} RENAME TO {self.table_name}")
            return

        if mode == "append":
            self.conn.execute(
                f"INSERT INTO {self.table_name} SELECT * FROM {temp_name}"
            )

        elif mode == "merge":
            # Insert only rows where key columns don't exist in target
            key_matches = " AND ".join(f"t.{col} = m.{col}" for col in key_columns)
            self.conn.execute(
                f"""
                INSERT INTO {self.table_name}
                SELECT t.* FROM {temp_name} t
                WHERE NOT EXISTS (
                    SELECT 1 FROM {self.table_name} m
                    WHERE {key_matches}
                )
            """
            )

        elif mode == "update":
            # Create staging table
            staging_name = f"staging_{self.table_name}"

            # First copy all rows from target that don't match key columns
            key_matches = " AND ".join(f"t.{col} = m.{col}" for col in key_columns)
            self.conn.execute(
                f"""
                CREATE TABLE {staging_name} AS
                SELECT t.* FROM {self.table_name} t
                WHERE NOT EXISTS (
                    SELECT 1 FROM {temp_name} m
                    WHERE {key_matches}
                )
            """
            )

            # Then add all rows from new data
            self.conn.execute(
                f"""
                INSERT INTO {staging_name}
                SELECT * FROM {temp_name}
            """
            )

            # Replace original with staging
            self.conn.execute(f"DROP TABLE {self.table_name}")
            self.conn.execute(f"ALTER TABLE {staging_name} RENAME TO {self.table_name}")

        # Cleanup
        self.conn.execute(f"DROP TABLE IF EXISTS {temp_name}")



# from datetime import datetime
# from typing import Any, Dict, List, Optional

# from pydantic import BaseModel


# class SchedulerBase(BaseModel):
#     name: str
#     executor: Optional[str] = "local"
#     inputs: Optional[Dict[str, Any]] = None
#     final_vars: Optional[List[str]] = None
#     config: Optional[Dict[str, Any]] = None
#     with_tracker: bool = False
#     with_opentelemetry: bool = False


# class SchedulerAdd(SchedulerBase):
#     type: str = "cron"
#     paused: bool = False
#     coalesce: str = "latest"
#     misfire_grace_time: Optional[float] = None
#     max_jitter: Optional[float] = None
#     max_running_jobs: Optional[int] = None
#     conflict_policy: str = "do_nothing"
#     crontab: Optional[str] = None
#     cron_params: Optional[Dict[str, Any]] = None
#     interval_params: Optional[Dict[str, Any]] = None
#     calendarinterval_params: Optional[Dict[str, Any]] = None
#     date_params: Optional[Dict[str, Any]] = None
#     storage_options: Optional[Dict[str, Any]] = None


# class SchedulerModify(BaseModel):
#     name: str
#     paused: Optional[bool] = None
#     next_run_time: Optional[datetime] = None


# class SchedulerDelete(BaseModel):
#     name: str


# class SchedulerList(BaseModel):
#     pattern: Optional[str] = None


# class SchedulerInfo(BaseModel):
#     name: str



from ..base import BaseFileLoader


class JsonFileLoader(BaseFileLoader):
    """
    JSON file loader.

    This class is responsible for loading dataframes from JSON files.

    Examples:
        ```python
        loader = JsonFileLoader("data.json")
        df = loader.load()
        ```
    """

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "json"


class JsonDatasetLoader(BaseFileLoader):
    """
    JSON dataset loader.

    This class is responsible for loading dataframes from JSON dataset.

    Examples:
        ```python
        loader = JsonDatasetLoader("json_data/")
        df = loader.load()
        ```
    """

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "json"



# from ..cli.utils import parse_dict_or_list_param
from pydantic import BaseModel, ValidationError as PydandticValidationError
from sanic_ext.exceptions import ValidationError

from typing import Any


async def deserialize_and_validate(
    type_: BaseModel,
    body: dict | None = None,
    query: list[tuple[str, Any]] | None = None,
):
    try:
        if query:
            return type_(**dict(query) or {})
        return type_(**body or {})
    except (TypeError, PydandticValidationError) as e:
        raise ValidationError(e)
